﻿using PalmSens;
using PalmSens.Comm;
using PalmSens.Core.Simplified.Data;
using PalmSens.Devices;
using PalmSens.Techniques;
using System;
using System.Threading.Tasks;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

// The Blank Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=402352&clcid=0x409

namespace PSSDKBasicExample
{
    /// <summary>
    /// An empty page that can be used on its own or navigated to within a Frame.
    /// </summary>
    public sealed partial class MainPage : Page
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="MainPage"/> class.
        /// </summary>
        public MainPage()
        {
            this.InitializeComponent();
        }

        /// <summary>
        /// Handles the Loaded event of the Page control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void Page_Loaded(object sender, RoutedEventArgs e)
        {
            await DiscoverConnectedDevices();
            InitResultsListBox();
            InitLSVMethod();
        }

        /// <summary>
        /// The instance of method class containing the Linear Sweep Voltammetry parameters
        /// </summary>
        private LinearSweep _methodLSV;

        /// <summary>
        /// The connected PalmSens & EmStat devices
        /// </summary>
        private Device[] _connectedDevices = new Device[0];

        /// <summary>
        /// The active SimpleMeasurement
        /// </summary>
        private SimpleMeasurement _activeMeasurement = null;

        /// <summary>
        /// The active SimpleCurve
        /// </summary>
        private SimpleCurve _activeCurve = null;

        /// <summary>
        /// Initializes the listbox control.
        /// </summary>
        private void InitResultsListBox()
        {
            lbResults.Items.Clear();
            lbResults.Items.Add("ID     Potential (V)   Current (µA)");
        }

        /// <summary>
        /// Initializes the LSV method.
        /// </summary>
        private void InitLSVMethod()
        {
            _methodLSV = new LinearSweep(); //Create a new linear sweep method with the default settings
            _methodLSV.BeginPotential = -.5f; //Sets the potential to start the sweep from
            _methodLSV.EndPotential = .5f - 0.05f; //Sets the potential for the sweep to stop at
            _methodLSV.StepPotential = 0.05f; //Sets the step size
            _methodLSV.Scanrate = 0.1f; //Sets the scan rate to 0.05 V/s

            _methodLSV.EquilibrationTime = 1f; //Equilabrates the cell at the defined potential for 1 second before starting the measurement
            _methodLSV.Ranging.StartCurrentRange = new CurrentRange(CurrentRanges.cr1uA); //Starts equilabration in the 1µA current range
            _methodLSV.Ranging.MinimumCurrentRange = new CurrentRange(CurrentRanges.cr10nA); //Min current range 10nA
            _methodLSV.Ranging.MaximumCurrentRange = new CurrentRange(CurrentRanges.cr1mA); //Max current range 1mA
        }

        /// <summary>
        /// Discovers the connected PalmSens & EmStat devices and adds them to the combobox control.
        /// </summary>
        private async Task DiscoverConnectedDevices()
        {
            btnRefresh.IsEnabled = false;
            cmbDevices.Items.Clear();
            _connectedDevices = await psCommSimpleUWP.GetConnectedDevices(); //Discover connected devices

            foreach (Device d in _connectedDevices)
                cmbDevices.Items.Add(d.ToString()); //Add connected devices to control

            int nDevices = cmbDevices.Items.Count;
            cmbDevices.SelectedIndex = nDevices > 0 ? 0 : -1;
            lbConsole.Items.Add($"Found {nDevices} device(s).");

            btnConnect.IsEnabled = nDevices > 0;
            btnRefresh.IsEnabled = true;
        }

        /// <summary>
        /// Handles the Click event of the btnRefresh control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void btnRefresh_Click(object sender, RoutedEventArgs e)
        {
            await DiscoverConnectedDevices();
        }

        /// <summary>
        /// Handles the Click event of the btnConnect control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void btnConnect_Click(object sender, RoutedEventArgs e)
        {
            btnConnect.IsEnabled = false;
            if (!psCommSimpleUWP.Connected) //Determine whether a device is currently connected
            {
                if (cmbDevices.SelectedIndex == -1)
                    return;

                try
                {
                    //Connect to the device selected in the devices combobox control
                    await psCommSimpleUWP.Connect(_connectedDevices[cmbDevices.SelectedIndex]);
                    lbConsole.Items.Add($"Connected to {psCommSimpleUWP.ConnectedDevice.ToString()}");
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            else
            {
                await psCommSimpleUWP.Disconnect(); //Disconnect from the connected device
                lbConsole.Items.Add("Disconnected");
            }

            //Update UI based on connection status
            cmbDevices.IsEnabled = !psCommSimpleUWP.Connected;
            btnRefresh.IsEnabled = !psCommSimpleUWP.Connected;
            btnConnect.Content = psCommSimpleUWP.Connected ? "Disconnect" : "Connect";
            btnMeasure.IsEnabled = psCommSimpleUWP.Connected;
            btnConnect.IsEnabled = true;
        }

        /// <summary>
        /// Handles the Click event of the btnMeasure control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void btnMeasure_Click(object sender, RoutedEventArgs e)
        {
            btnMeasure.IsEnabled = false;
            if (psCommSimpleUWP.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    InitResultsListBox(); //Reset the data list box
                    _activeMeasurement = await psCommSimpleUWP.Measure(_methodLSV); //Start measurement defined in the method
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            else
            {
                try
                {
                    await psCommSimpleUWP.AbortMeasurement(); //Abort the active measurement
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            btnMeasure.IsEnabled = true;
        }

        /// <summary>
        /// Raised when the measurement is ended
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void psCommSimpleUWP_MeasurementEnded(object sender, EventArgs e)
        {
            int nDataPointsReceived = _activeCurve != null ? _activeCurve.NDataPoints : 0;
            lbConsole.Items.Add($"{nDataPointsReceived} data point(s) received.");
            lbConsole.Items.Add("Measurement ended.");
        }

        /// <summary>
        /// Raised when the measurement is started
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void psCommSimpleUWP_MeasurementStarted(object sender, EventArgs e)
        {
            lbConsole.Items.Add("Linear sweep voltammetry measurement started.");
        }

        /// <summary>
        /// Raised when device status package is received (the device does not send status packages while measuring)
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="PalmSens.Comm.StatusEventArgs"/> instance containing the event data.</param>
        private void psCommSimpleUWP_ReceiveStatus(object sender, PalmSens.Comm.StatusEventArgs e)
        {
            Status status = e.GetStatus(); //Get the PalmSens.Comm.Status instance from the event data
            double potential = status.PotentialReading.Value; //Get the potential
            double currentInRange = status.CurrentReading.ValueInRange; //Get the current expressed inthe active current range
            PalmSens.Comm.ReadingStatus currentStatus = status.CurrentReading.ReadingStatus; //Get the status of the current reading
            CurrentRange cr = status.CurrentReading.CurrentRange; //Get the active current range

            tbPotential.Text = potential.ToString("F3");
            tbCurrent.Text = currentInRange.ToString("F3");
            lblCurrentRange.Text = $"* {cr.ToString()}";
        }

        /// <summary>
        /// Raised when the connected device's status changes
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="CurrentState">State of the current.</param>
        private void psCommSimpleUWP_StateChanged(object sender, PalmSens.Comm.CommManager.DeviceState CurrentState)
        {
            tbStatus.Text = CurrentState.ToString(); //Updates the device state indicator textbox
            btnConnect.IsEnabled = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle;
            btnMeasure.Content = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle ? "Measure" : "Abort";
        }

        /// <summary>
        /// Raised when a Simple Curve in the active SimpleMeasurement starts receiving data
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="activeSimpleCurve">The active simple curve.</param>
        private void psCommSimpleUWP_SimpleCurveStartReceivingData(object sender, SimpleCurve activeSimpleCurve)
        {
            _activeCurve = activeSimpleCurve; //Get the reference to the active SimpleCurve

            //Subscribe to the curve's events to receive updates when new data is available and when it iss finished receiving data
            _activeCurve.NewDataAdded += _activeCurve_NewDataAdded;
            _activeCurve.CurveFinished += _activeCurve_CurveFinished;

            lbConsole.Items.Add("Curve is receiving new data...");
        }

        /// <summary>
        /// Raised when new data points are added to the active SimpleCurve
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="PalmSens.Data.ArrayDataAddedEventArgs"/> instance containing the event data.</param>
        private void _activeCurve_NewDataAdded(object sender, PalmSens.Data.ArrayDataAddedEventArgs e)
        {
            if (!Dispatcher.HasThreadAccess) //Data is parsed asynchronously in the case this event was raised on a different thread it must be invoked back onto the UI thread
            {
                Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () => _activeCurve_NewDataAdded(sender, e)); //Recast event to UI thread
                return;
            }
            int startIndex = e.StartIndex; //The index of the first new data point added to the curve
            int count = e.Count; //The number of new data points added to the curve

            for (int i = startIndex; i < startIndex + count; i++)
            {
                double xValue = _activeCurve.XAxisValue(i); //Get the value on Curve's X-Axis (potential) at the specified index
                double yValue = _activeCurve.YAxisValue(i); //Get the value on Curve's Y-Axis (current) at the specified index
                lbResults.Items.Add($"{i+1}     {xValue.ToString("F2")}     {yValue.ToString("E3")}");
            }

            tbPotential.Text = _activeCurve.XAxisValue(startIndex + count - 1).ToString("F3");
            tbCurrent.Text = _activeCurve.YAxisValue(startIndex + count - 1).ToString("F3");
        }

        /// <summary>
        /// Raised when a SimpleCurve stops receiving new data points
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void _activeCurve_CurveFinished(object sender, EventArgs e)
        {
            if (!Dispatcher.HasThreadAccess) //Data is parsed asynchronously in the case this event was raised on a different thread it must be invoked back onto the UI thread
            {
                Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () => _activeCurve_CurveFinished(sender, e)); //Recast event to UI thread
                return;
            }
            //Unsubscribe from the curves events to avoid memory leaks
            _activeCurve.NewDataAdded -= _activeCurve_NewDataAdded;
            _activeCurve.CurveFinished -= _activeCurve_CurveFinished;

            lbConsole.Items.Add("Curve Finished");
        }
    }
}
